
#include "procedural.hpp"
#include "model.hpp"

using namespace fggl::data;

fggl::data::Mesh fggl::data::make_triangle() {
    constexpr fggl::math::vec3 pos[]{
            {-0.5f, -0.5f, 0.0f},
            {0.5f,  -0.5f, 0.0f},
            {0.0f,  0.5f,  0.0f}
    };

    // add points
    fggl::data::Mesh mesh;
    for (auto po : pos) {
        Vertex vert{};
        vert.posititon = po;
        vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
        mesh.push(vert);
    }

    // mesh
    for (int i = 0; i < 3; ++i) {
        mesh.pushIndex(i);
    }

    return mesh;
}

fggl::data::Mesh fggl::data::make_quad_xy() {
	constexpr fggl::math::vec3 pos[] {
		{-0.5f, -0.5f, 0.0f},
		{ 0.5f, -0.5f, 0.0f},
		{ 0.5f,  0.5f, 0.0f},
		{-0.5f,  0.5f, 0.0f}
	};
	constexpr int idx[] {
		0, 1, 3,
		3, 1, 2
	};

	fggl::data::Mesh mesh;
	int colIdx[4];
	for (int i = 0; i < 4; ++i){
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex(vert);
	}

	for( auto i : idx ) {
		mesh.pushIndex( colIdx[ i ] );
	}

	return mesh;
}

fggl::data::Mesh fggl::data::make_quad_xz() {
	constexpr fggl::math::vec3 pos[] {
		{-0.5f, 0.0f, -0.5f},
		{ 0.5f, 0.0f, -0.5f},
		{ 0.5f, 0.0f,  0.5f},
		{-0.5f, 0.0f,  0.5f}
	};
	constexpr int idx[] {
		0, 1, 3,
		3, 1, 2
	};

	fggl::data::Mesh mesh;
	int colIdx[4];
	for (int i = 0; i < 4; ++i){
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex(vert);
	}

	for( auto i : idx ) {
		mesh.pushIndex( colIdx[ i ] );
	}

	return mesh;
}


fggl::data::Mesh fggl::data::make_cube() {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr int idx[] {
		0, 3, 1, // top
		3, 2, 1,
		0, 1, 4, // side 0 - 1
		5, 4, 1,
		1, 2, 5, // side 1 - 2
		2, 6, 5,
		3, 7, 2, // side 2 - 3
		2, 7, 6,
		0, 4, 3, // side 3 - 0
		4, 7, 3,
		4, 5, 7, // bottom
		7, 5, 6,
	};

	fggl::data::Mesh mesh;
	int colIdx[8];
	for ( int i=0; i < 8; ++i ) {
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex( vert );
	}

	for ( auto i : idx ){
		mesh.pushIndex( colIdx[i] );
	}

	return mesh;
}

fggl::data::Mesh fggl::data::make_slope() {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	// FIXME remove 2 and 3 and renumber the index list accordingly
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr int idx[] {
		0, 7, 1, // ramp
		7, 6, 1,
		0, 1, 4, // side 0 - 1
		5, 4, 1,
		1, 6, 5, // ramp side 1
		4, 7, 0, // ramp side 2
		4, 5, 7, // bottom
		7, 5, 6,
	};

	fggl::data::Mesh mesh;
	int colIdx[8];
	for ( int i=0; i < 8; ++i ) {
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex( vert );
	}

	for ( auto i : idx ){
		mesh.pushIndex( colIdx[i] );
	}

	return mesh;
}

fggl::data::Mesh fggl::data::make_ditch() {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	// FIXME remove 2 and renumber the index list accordingly
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr int idx[] {
		0, 3, 1, // top
		3, 6, 1,
		0, 1, 4, // side 0 - 1
		5, 4, 1,
		1, 6, 5,
		3, 7, 6,
		0, 4, 3, // side 3 - 0
		4, 7, 3,
		4, 5, 7, // bottom
		7, 5, 6,
	};

	fggl::data::Mesh mesh;
	int colIdx[8];
	for ( int i=0; i < 8; ++i ) {
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex( vert );
	}

	for ( auto i : idx ){
		mesh.pushIndex( colIdx[i] );
	}

	return mesh;
}

fggl::data::Mesh fggl::data::make_point() {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr int idx[] {
		0, 7, 5, // top
		7, 6, 5,
		0, 5, 4, // side 0 - 1
	//	5, 4, 1,
	//	1, 2, 5, // side 1 - 2
	//	2, 6, 5,
	//	3, 7, 2, // side 2 - 3
	//	2, 7, 6,
		0, 4, 7, // side 3 - 0
	//	4, 7, 3,
		4, 5, 7, // bottom
		7, 5, 6,
	};

	fggl::data::Mesh mesh;
	int colIdx[8];
	for ( int i=0; i < 8; ++i ) {
		Vertex vert{};
		vert.posititon = pos[i];
		vert.colour = fggl::math::vec3(1.0f, 1.0f, 1.0f);
		colIdx[ i ] = mesh.pushVertex( vert );
	}

	for ( auto i : idx ){
		mesh.pushIndex( colIdx[i] );
	}

	return mesh;
}



/*
constexpr float cubeVertex[8 * 3] = {
	0, 0, 0,
	0, 1, 0,
	1, 1, 0,
	1, 0, 0,
	0, 0, 1,
	0, 1, 1,
	1, 1, 1,
	0, 0, 0,
};

constexpr int cubeIndexes[6 * 4] = {
	7, 6, 5, 4,
	0, 1, 2, 3,
	6, 7, 3, 2,
	5, 6, 2, 1,
	4, 5, 1, 0,
	7, 4, 0, 3,
};


void quadsToTris(fggl::data::Mesh& data, const std::vector<unsigned int>& idx, const int* quads, const int nQuads) {
	for (int p = 0; p < nQuads; p++) {
		int offset = p * 4;
		// first tri
		data.pushIndex( idx[ quads[offset + 0 ] ] );
		data.pushIndex( idx[ quads[offset + 1 ] ] );
		data.pushIndex( idx[ quads[offset + 2 ] ] );

		// second tri
		data.pushIndex( idx[ quads[offset + 2 ] ] );
		data.pushIndex( idx[ quads[offset + 3 ] ] );
		data.pushIndex( idx[ quads[offset + 0 ] ] );
	}
}

inline glm::vec3 calcNormal(Vertex prev, Vertex next, Vertex curr) {
	return glm::cross( prev.posititon - curr.posititon, next.posititon - curr.posititon );
}

void computeNormals(Mesh& mesh) {

	const int idxes = mesh.indexCount();
	const int vertCount = mesh.vertexCount();

	// zero the normals
	for (int i=0; i<vertCount; i++) {
		Vertex& vert = mesh.vertex( i );
		vert.normal = glm::vec3(0.0f);
	}

	// calculate normals for all triangles
	for (int tri=0; tri<idxes; tri += 3) {
		// get next 3 points
		Vertex& pa = mesh.vertex( tri );
		Vertex& pb = mesh.vertex( tri + 1 );
		Vertex& pc = mesh.vertex( tri + 2 );

		pa.normal += calcNormal( pb, pc, pa );
		pb.normal += calcNormal( pc, pa, pb );
		pc.normal += calcNormal( pa, pb, pc );
	}

	// normalise
	for (int i=0; i<vertCount; i++) {
		Vertex& vert = mesh.vertex( i );
		vert.normal = glm::normalize( vert.normal );
	}

}

Mesh fggl::data::make_cube() {

	int nVerts = sizeof(cubeVertex) / sizeof(cubeVertex[0]) / 3;
	int nQuads = sizeof(cubeIndexes) / sizeof(cubeIndexes[0]) / 4;

	std::vector< unsigned int > vertIndex;

	Mesh cube;

	// stick the vertex data into the mesh
	for ( int i=0; i < nVerts; i++ ) {
		fggl::data::Vertex vert;
		vert.posititon = fggl::math::vec3( cubeVertex[i], cubeVertex[i + 1], cubeVertex[i+2] );

		auto idx = cube.pushVertex(vert);
		vertIndex.push_back(idx);
	}

	// triagulate the quads
	quadsToTris(cube, vertIndex, cubeIndexes, nQuads);
	computeNormals(cube);

	return cube;
}*/

